#include <vtk.h>
#include <cave_ogl.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <vtkCAVERenderWindow.h>
#include <vtkCAVERenderer.h>
#include <vtkMarchingCubes.h>
#include <vtkDecimate.h>
//#include <vtkStructuredPointsReader.h>
//#include <vtkStructuredPointsToPolyDataFilter.h>

#ifndef DTOR
#define DTOR(r) r*(M_PI/180.0)
#endif

// The vtkCAVE variables
vtkCAVERenderWindow* caveRenderWindow;
vtkCAVERenderer* caveRenderer;

// Note vtk objects can NOT be put into
// shared memory, hence the reason for global
// distributed memory objects.
vtkActor* cylinderActor;
vtkMatrix4x4* cylinderMatrix;
vtkTransform* caveTransform;
vtkTransform* cave2Actor;
vtkPlane* plane;
vtkActor* cutActor;
vtkMaskPoints *maskPoints1, *maskPoints2, *maskPoints3;
int *maskOffset;

// Function prototypes
void draw();
void frameUpdate();
void createCombustor(vtkStructuredPointsReader*);
void getWandMatrix(vtkMatrix4x4*);
void initVTK(int, char**);
void navTranslate(float);
void CAVENavRotateY();
void navigate();


// This is the draw function that gets passed to the CAVELib
// All rendering should be done in this function between the 
// glPushMatrix() and glPopMatrix() calls.
void
draw()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClearColor(0,0,0,0);

    glPushMatrix();

       // Get the cave Navigation Transformation.  What allows the user
       // to move in the CAVE(with wand after a little help from us in
       // the frame function, navigate function call)
       CAVENavTransform();
       caveRenderWindow->Render();

    glPopMatrix();
}


// This frame function gets called once per frame
// and handles user interactions.
void
frameUpdate(){

    //float pos[4], *posInActor;
     
    if(CAVEBUTTON2){  // give life to the glyphs
          maskPoints1->SetOffset((*maskOffset)++);
          maskPoints2->SetOffset((*maskOffset)++);
          maskPoints3->SetOffset((*maskOffset)++);
          if(*maskOffset == 29)
             *maskOffset = -1; // reset to the beginning offset
    }

    //
    // This code was used for making this example.  Pressing the
    // m key on the keyboard prints out the position of the wand
    // This was used to determine where to place the starting points
    // for the stream points
    // 
    //pos[3] = 1;
    //if(CAVEgetbutton(CAVE_MKEY))
    //{
        //CAVEGetPosition(CAVE_WAND_NAV,pos);
        //cave2Actor->SetPoint(pos);
        //posInActor = cave2Actor->GetPoint();
        //fprintf(stderr, "%f2., %f.2, %f.2\n", posInActor[0], posInActor[1], posInActor[2]);
    //}

    if(CAVEBUTTON3){  // allow resizing of the cumbustion chamber
       // pressing the pad on the wand is the joystick input
       if(CAVE_JOYSTICK_Y > 0.5)
       {
          caveTransform->Translate(0.0, -5.0, 0.0);
          caveTransform->Scale(1.025, 1.025, 1.025);
          caveTransform->Translate(0.0, 5.0, 0.0);
       }if(CAVE_JOYSTICK_Y < -0.5)
       {
          caveTransform->Translate(0.0, -5.0, 0.0);
          caveTransform->Scale(0.975, 0.975, 0.975);
          caveTransform->Translate(0.0, 5.0, 0.0);
       }
    }
    else
       // navigate by the pad on the wand.
       navigate(); 


    // Get the transformation matrix of the wand and set the cylinder
    // representing the wand to have the same transformation
    getWandMatrix(cylinderMatrix);
    cylinderActor->SetUserMatrix(cylinderMatrix);

}


// Initialize the vtk objects
void
initVTK(int argc, char** argv)
{
    float pos[] = {-5, 0, 5};

    // Create the necessary vtkCAVE objects
    caveRenderer = vtkCAVERenderer::New();
    caveRenderWindow = vtkCAVERenderWindow::New();

    caveRenderWindow->AddRenderer(caveRenderer);
    caveRenderer->TwoSidedLightingOn();

    // Create some lights
    vtkLight* light1 = vtkLight::New();
    vtkLight* light2 = vtkLight::New();
    vtkLight* light3 = vtkLight::New();
    vtkLight* light4 = vtkLight::New();

    light1->SetPosition(pos[0], pos[1], pos[2]);
    light1->SetConeAngle(180);
    light1->SetColor(1.0, 1.0, 1.0);

    pos[0] = 5;
    light2->SetPosition(pos[0], pos[1], pos[2]);
    light2->SetConeAngle(180);
    light2->SetColor(1.0, 1.0, 1.0);

    pos[0] = 0;
    pos[1] = 10;
    pos[2] = 3;
    light3->SetPosition(pos[0],pos[1],pos[2]);
    light3->SetConeAngle(180);
    light3->SetColor(1,1,1);

    pos[0] = 0;
    pos[1] = 0;
    pos[2] = 0;
    light4->SetPosition(pos[0],pos[1],pos[2]);
    light4->SetConeAngle(180);
    light4->SetFocalPoint(0.0, 5.0, 0.0);
    light4->SetColor(1,1,1);

    caveRenderer->AddLight(light1);
    caveRenderer->AddLight(light2);
    caveRenderer->AddLight(light3);
    caveRenderer->AddLight(light4);

    // Create the reader
    vtkStructuredPointsReader *reader = vtkStructuredPointsReader::New();

    if(argc > 1){ // there is a arguement telling the 
          // directory where the .bin files are
       char *dataSetLocation = new char[80];
       sprintf(dataSetLocation, "%s/carotid.vtk", argv[1]);
       reader->SetFileName(dataSetLocation);
       delete(dataSetLocation);
    }else{ // no arguement assuming current directory
       reader->SetFileName("carotid.vtk");
    }
    reader->Update();

    // Create the Cumbustor and cutting Plane
    createCombustor(reader);
}

// create the combustion chamber outline and the cutting plane
void
createCombustor(vtkStructuredPointsReader* reader)
{
    float *center;

    // find the center point of the data.  Is used for making
    // the caveTransformation(i.e. putting the data into a
    // logical size and location for the CAVE)
    center = reader->GetOutput()->GetCenter();

    caveTransform = vtkTransform::New();
      // Translate to center of CAVE
      caveTransform->Translate(0.0, 5.0, 0.0);
      caveTransform->Scale(0.125, 0.125, 0.125);
      // translate to the center(ie 0,0,0) for scale
      caveTransform->Translate(-center[0], -center[1], -center[2]);

     // make a reverse transformation for putting coordinate values
     // into the data(Actor) coordinates
     cave2Actor = vtkTransform::New();
      cave2Actor->SetMatrix(*caveTransform->GetMatrixPointer());
      cave2Actor->Inverse();
      cave2Actor->PostMultiply();
    
    // Create a cylinder that will represent the ray from
    // the front of the wand
    vtkCylinderSource* cylinder = vtkCylinderSource::New();
    cylinder->SetRadius(0.05);
    cylinder->SetHeight(5.0);

    // Map the objects to graphics primitives
    vtkDataSetMapper* cylinderMapper = vtkDataSetMapper::New();
    cylinderMapper->SetInput(cylinder->GetOutput());

    cylinderActor = vtkActor::New();
    cylinderActor->SetMapper(cylinderMapper);
    cylinderActor->SetPosition(0,0,-2.5);
    cylinderActor->RotateX(90.0);
    cylinderActor->PickableOff();

    // Create a global matrix object to be used
    // in getting the orientation of the wand and
    // setting the ray cylinder to align with the wand
    cylinderMatrix = vtkMatrix4x4::New();

    // add the wand pointer
    caveRenderer->AddActor(cylinderActor);

//
// Outline of the area
//
  vtkOutlineFilter *outlineF = vtkOutlineFilter::New();
    outlineF->SetInput(reader->GetOutput());
  vtkPolyDataMapper *outlineMapper = vtkPolyDataMapper::New();
    outlineMapper->SetInput(outlineF->GetOutput());
  vtkActor *outline = vtkActor::New();
    outline->SetMapper(outlineMapper);
    outline->GetProperty()->SetColor(1,1,1);
    outline->SetUserMatrix(caveTransform->GetMatrixPointer());

//
// create the carotid isosurface
//

    vtkContourFilter *iso = vtkContourFilter::New();
       iso->SetInput(reader->GetOutput());
       iso->SetValue( 0,  190);
    vtkDecimate *decimate= vtkDecimate::New();
       decimate->SetInput(iso->GetOutput());
       decimate->SetTargetReduction(0.9);
    vtkPolyDataMapper *isoMapper = vtkPolyDataMapper::New();
       isoMapper->SetInput(decimate->GetOutput());
       isoMapper->ScalarVisibilityOff();
    vtkActor *isoActor = vtkActor::New();
       isoActor->SetMapper(isoMapper);
       isoActor->GetProperty()->SetRepresentationToWireframe();
       isoActor->GetProperty()->SetOpacity(0.005);
       isoActor->GetProperty()->SetColor(.5, .5, .5);
       isoActor->SetUserMatrix(caveTransform->GetMatrixPointer());
 
//
// Create the glyphs which represent the flow through the arteries
// 
// To create the illusion of watching the blood flow, we use StreamPoints
// to create a series of points along the flow from determined starting
// points.  By using the Mask we only display a portion of the glyphs
// representing the flow.  This allows us to cycle through the set(using
// the offset and create an illusion of particles being released into 
// the blood stream in real time.

   // make the glyph representation
    vtkConeSource *cone = vtkConeSource::New();
       cone->SetResolution(3);
       cone->SetHeight(1);
       cone->SetRadius(0.25);

   // First release point
   vtkPointSource *points1 = vtkPointSource::New();
      points1->SetNumberOfPoints(25);
      points1->SetCenter(134.06, 117.68, 4.72);
      points1->SetRadius(3);
   vtkStreamPoints *streamPoints1 = vtkStreamPoints::New();
      streamPoints1->SetInput(reader->GetOutput());
      streamPoints1->SetSource(points1->GetOutput());
      streamPoints1->SetTimeIncrement(.1);
      streamPoints1->Update();
   maskPoints1 = vtkMaskPoints::New();
      maskPoints1->SetInput(streamPoints1->GetOutput());
      maskPoints1->SetOnRatio(30);
      maskPoints1->SetOffset(0);
   vtkGlyph3D *streamCones1 = vtkGlyph3D::New();
      streamCones1->SetInput(maskPoints1->GetOutput());
      streamCones1->SetSource(cone->GetOutput());
      streamCones1->SetScaleFactor(0.5);
      streamCones1->SetScaleModeToScaleByVector();
   vtkLookupTable* lut = vtkLookupTable::New();
       lut->SetHueRange(.667, 0.0);
       lut->Build();
   vtkPolyDataMapper *streamMapper1 = vtkPolyDataMapper::New();
      streamMapper1->SetInput(streamCones1->GetOutput());
      streamMapper1->SetScalarRange(2, 7.5);
      streamMapper1->SetLookupTable(lut);
    vtkActor *streams1 = vtkActor::New();
      streams1->SetMapper(streamMapper1);
      streams1->SetUserMatrix(caveTransform->GetMatrixPointer());

   // Second release point
   vtkPointSource *points2 = vtkPointSource::New();
      points2->SetNumberOfPoints(25);
      points2->SetCenter(169.22, 117.68, 5.6);
      points2->SetRadius(2);
   vtkStreamPoints *streamPoints2 = vtkStreamPoints::New();
      streamPoints2->SetInput(reader->GetOutput());
      streamPoints2->SetSource(points2->GetOutput());
      streamPoints2->SetTimeIncrement(.1);
      streamPoints2->Update();
   maskPoints2 = vtkMaskPoints::New();
      maskPoints2->SetInput(streamPoints2->GetOutput());
      maskPoints2->SetOnRatio(30);
      maskPoints2->SetOffset(0);
   vtkGlyph3D *streamCones2 = vtkGlyph3D::New();
      streamCones2->SetInput(maskPoints2->GetOutput());
      streamCones2->SetSource(cone->GetOutput());
      streamCones2->SetScaleFactor(0.5);
      streamCones2->SetScaleModeToScaleByVector();
   vtkPolyDataMapper *streamMapper2 = vtkPolyDataMapper::New();
      streamMapper2->SetInput(streamCones2->GetOutput());
      streamMapper2->SetScalarRange(2, 7.5);
      streamMapper2->SetLookupTable(lut);
    vtkActor *streams2 = vtkActor::New();
      streams2->SetMapper(streamMapper2);
      streams2->SetUserMatrix(caveTransform->GetMatrixPointer());

    // third release point
   vtkPointSource *points3 = vtkPointSource::New();
      points3->SetNumberOfPoints(25);
      points3->SetCenter(116.6, 101.3, 4.5);
      points3->SetRadius(3);
   vtkStreamPoints *streamPoints3 = vtkStreamPoints::New();
      streamPoints3->SetInput(reader->GetOutput());
      streamPoints3->SetSource(points3->GetOutput());
      streamPoints3->SetTimeIncrement(.1);
      streamPoints3->Update();
   maskPoints3 = vtkMaskPoints::New();
      maskPoints3->SetInput(streamPoints3->GetOutput());
      maskPoints3->SetOnRatio(30);
      maskPoints3->SetOffset(0);
   vtkGlyph3D *streamCones3 = vtkGlyph3D::New();
      streamCones3->SetInput(maskPoints3->GetOutput());
      streamCones3->SetSource(cone->GetOutput());
      streamCones3->SetScaleFactor(0.5);
      streamCones3->SetScaleModeToScaleByVector();
   vtkPolyDataMapper *streamMapper3 = vtkPolyDataMapper::New();
      streamMapper3->SetInput(streamCones3->GetOutput());
      streamMapper3->SetScalarRange(2, 7.5);
      streamMapper3->SetLookupTable(lut);
    vtkActor *streams3 = vtkActor::New();
      streams3->SetMapper(streamMapper3);
      streams3->SetUserMatrix(caveTransform->GetMatrixPointer());

    caveRenderer->AddActor(outline);   
    caveRenderer->AddActor(streams1);   
    caveRenderer->AddActor(streams2);   
    caveRenderer->AddActor(streams3);   
    caveRenderer->AddActor(isoActor);   

    // set up for movement of the cutting plane
    maskOffset = (int*) CAVEMalloc(sizeof(int));
    *maskOffset = 0;

    // set up for resizing of the cumbustor
    caveTransform->PostMultiply();
}

void
main(int argc, char **argv)
{

    CAVEConfigure(&argc,argv,NULL);
    CAVEInit();
    CAVEInitApplication((CAVECALLBACK)initVTK, 2, argc, argv);
    CAVEFrameFunction(frameUpdate, 0);
    CAVEDisplay(draw, 0);
    
    while(!CAVEgetbutton(CAVE_ESCKEY))
    {
          
    }

    // free up the memory we allocated
    CAVEFree(maskOffset);
    
    CAVEExit();

    caveRenderer->Delete();
    caveRenderWindow->Delete();
}

// From the wands position and orientation determine
// its transformation matrix.
void
getWandMatrix(vtkMatrix4x4* m)
{
    vtkTransform *transform;
    float or[3],pos[3];

    transform = vtkTransform::New();

    CAVEGetPosition(CAVE_WAND_NAV,pos);
    CAVEGetOrientation(CAVE_WAND_NAV,or);

    // note the order of rotation.  This is very important
    transform->Translate(pos[0], pos[1], pos[2]);
    transform->RotateY(or[1]);
    transform->RotateX(or[0]);
    transform->RotateZ(or[2]);

    m->DeepCopy(transform->GetMatrixPointer());
    transform->Delete();

}

// Handles the translation component of navigation
// based on the orientation of the wand
void
navTranslate(float dz)
{

    float wandOrn[3];
    float wx, wz;

    CAVEGetOrientation(CAVE_WAND,wandOrn);
    wx = dz * sinf(DTOR(wandOrn[1]))*5;
    wz = dz * cosf(DTOR(wandOrn[1]))*5;
    CAVENavTranslate(wx, 0, wz);

}


// Handles the rotation component of navigation
// based on the amount the joystick is pressed
void
CAVENavRotateY()
{
    float azi;

    azi = 1.0;

    if(CAVE_JOYSTICK_X > 0.5)
        CAVENavRot(-azi, 'y');

    if(CAVE_JOYSTICK_X < -0.5)
        CAVENavRot(azi, 'y');

}


// The basic navigation function. It calls the
// auxillary navigation functions.
void
navigate()
{

    float dz;

    if(CAVE_JOYSTICK_Y > 0.5)
    {
        dz = -.1;
        navTranslate(dz);
    }
    if(CAVE_JOYSTICK_Y < -0.5)
    {
        dz = .1;
        navTranslate(dz);
    }
    CAVENavRotateY();
}

