/* vtkCAVEPointPicker - Extends vtkPointPicker to work within CAVE environment
Inherits from vtkCAVEPicker, requires vtk library available from kitware
inc. <www.kitware.com>

This software has been developed as part of the Futures Laboratory
project at Argonne National Laboratory, for more information email
fl-info@mcs.anl.gov, and is distributed under the terms described at
http://www.mcs.anl.gov/fl/softwareterms.html

(c) 1998 Argonne National Laboratory/University of Chicago

*/

#include "vtkCAVEPointPicker.h"
#include "vtkMath.h"

vtkCAVEPointPicker::vtkCAVEPointPicker()
{
  this->PointId = -1;
}

float vtkCAVEPointPicker::IntersectWithLine(float p1[3], float p2[3], float tol, 
					vtkActor *assem, vtkActor *a, 
					vtkMapper *m)
{
  vtkDataSet *input=m->GetInput();
  int numPts;
  int ptId, i, minPtId;
  float ray[3], rayFactor, tMin, *p, t, projXYZ[3], minXYZ[3];

  if ( (numPts = input->GetNumberOfPoints()) < 1 )
    {
    return 2.0;
    }
//
//   Determine appropriate info
//
  for (i=0; i<3; i++)
    ray[i] = p2[i] - p1[i];

  if (( rayFactor = vtkMath::Dot(ray,ray)) == 0.0 ) 
    {
    vtkErrorMacro("Cannot process points");
    return 2.0;
    }
//
//  Project each point onto ray.  Keep track of the one within the
//  tolerance and closest to the eye (and within the clipping range).
//
  for (minPtId=(-1),tMin=VTK_LARGE_FLOAT,ptId=0; ptId<numPts; ptId++) 
    {
    p = input->GetPoint(ptId);

    t = (ray[0]*(p[0]-p1[0]) + ray[1]*(p[1]-p1[1]) + ray[2]*(p[2]-p1[2])) 
        / rayFactor;
//
//  If we find a point closer than we currently have, see whether it
//  lies within the pick tolerance and clipping planes.
//
    if ( t >= 0.0 && t <= 1.0 && t < tMin ) 
      {
      for(i=0; i<3; i++) 
        {
        projXYZ[i] = p1[i] + t*ray[i];
        if ( fabs(p[i]-projXYZ[i]) > tol )
	  {
	  break;
	  }
        }
      if ( i > 2 ) // within tolerance
        {
        minPtId = ptId;
        minXYZ[0]=p[0]; minXYZ[1]=p[1]; minXYZ[2]=p[2];
        tMin = t;
        }
      }
    }
//
//  Now compare this against other actors.
//
  if ( minPtId>(-1) && tMin < this->GlobalTMin ) 
    {
    this->MarkPicked(assem, a, m, tMin, minXYZ);
    this->PointId = minPtId;
    vtkDebugMacro("Picked point id= " << minPtId);
    }

  return tMin;
}

void vtkCAVEPointPicker::Initialize()
{
  this->PointId = (-1);
  this->vtkPicker::Initialize();
}

void vtkCAVEPointPicker::PrintSelf(ostream& os, vtkIndent indent)
{
  this->vtkPicker::PrintSelf(os,indent);

  os << indent << "Point Id: " << this->PointId << "\n";
}
