/* vtkCAVERenderWindow: Sets the vtk specific display parameters with
 CAVE display parameters for vtkCAVERenderer to use.  Inherits from
 vtkOpenGLRenderWindow, requires vtk library available from kitware
 inc. <www.kitware.com>

This software has been developed as part of the Futures Laboratory
project at Argonne National Laboratory, for more information email
fl-info@mcs.anl.gov, and is distributed under the terms described at
http://www.mcs.anl.gov/fl/softwareterms.html

(c) 1998 Argonne National Laboratory/University of Chicago

*/

#include <vtkCAVERenderWindow.h>
#include <cave_ogl.h>


// Originally gets called DEEP within vtkRenderWindow class
// virtually declared in vtkRenderWindow, and defined in vtkOpenGLRenderWindow
// class, if not overridden here the parent class method get's executed. This
// in turn causes the parent's Initialize() method to be executed, as
// oppose to the vtkCAVERenderWindow Initialize() method.
void vtkCAVERenderWindow::Start()
{

    if (!this->ContextId)
    {
	this->Initialize();
    }
    
    // set the current window 
    glXMakeCurrent(this->DisplayId,this->WindowId,this->ContextId);
  
}


// This method overides the parent's Initialize() method. In this way
// we set the WindowId, DisplayId, and ContextId, to those of the CAVE,
// replacing all the windowing calls that would've been done by the
// parent class, vtkOpenGLRenderWindow
void vtkCAVERenderWindow::Initialize()
{
    // Replaces the window calls within the parents Initialize() method
    this->WindowId = CAVEXWindow();
    this->DisplayId = CAVEXDisplay();
    this->ContextId = CAVEGLXContext();

    // Other initializations that normally take place in the
    // parent's Initialize() method
    vtkDebugMacro(<< " glMatrixMode ModelView\n");
    glMatrixMode( GL_MODELVIEW );
    
    vtkDebugMacro(<< " zbuffer enabled\n");
    glDepthFunc( GL_LEQUAL );
    glEnable( GL_DEPTH_TEST );

    vtkDebugMacro(" texture stuff\n");
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    // initialize blending for transparency
    vtkDebugMacro(<< " blend func stuff\n");
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    glEnable( GL_NORMALIZE );
    glAlphaFunc(GL_GREATER,0);

    // Don't think this is necessarry since the CAVE lib takes care of window
    // sizing, depending upon whether it's a sim or full CAVE
    // this->Mapped = 1;
    
}

// Overload the Frame function from vtkRenderWindow, because the parent methods
// calls a swap buffers. In vtkCAVE the CAVElibs is responsible for buffer swapping
void vtkCAVERenderWindow::Frame() {}
