/* vtkCAVERenderer: Calls necessary routines for rendering vtk objects
within CAVE.  Inherits from vtkOpenGLRenderer, requires vtk library
available from kitware inc. <www.kitware.com>

This software has been developed as part of the Futures Laboratory
project at Argonne National Laboratory, for more information email
fl-info@mcs.anl.gov, and is distributed under the terms described at
http://www.mcs.anl.gov/fl/softwareterms.html

(c) 1998 Argonne National Laboratory/University of Chicago

*/
#include <vtkCAVERenderer.h>
#include <GL/gl.h>


// Called by the Render() method. First pass thru it creates a Camera. This
// method overrides the parent method by removing the call to do camera
// based rendering.
int vtkCAVERenderer::UpdateCameras()
{
    if (!this->ActiveCamera)
    {
	vtkDebugMacro(<< "No cameras are on, creating one.");
	// the get method will automagically create a camera
	// and reset it since one hasn't been specified yet
	this->GetActiveCamera();
    }
    return 1;
}


// Overiding the base class Render() method. Calls methods to have volumes and actors
// render themselves. It also doesn't call some viewport resizing methods in order
// to accomodate varying volumes to be rendered.
//void vtkCAVERenderer::Render()
void vtkCAVERenderer::DeviceRender()
{
    int    actor_count;
    int    volume_count;

    if (this->StartRenderMethod) 
	(*this->StartRenderMethod)(this->StartRenderMethodArg);

    // standard render method 
    this->ClearLights();
    this->UpdateCameras();
    this->UpdateLights();
    actor_count =  this->UpdateActors();
    volume_count = this->UpdateVolumes();

    if ( !(actor_count + volume_count) )
	vtkWarningMacro(<< "No actors or volumes are on.");

    // Disable lighting turned on within UpdateLights()
    glDisable(GL_LIGHTING);

    // XXX removed 4-21-98, doesn't appear to be a problem
    // clean up the model view matrix set up by the camera 
//    glMatrixMode(GL_MODELVIEW);
//    glPopMatrix();

    if (this->EndRenderMethod) 
	(*this->EndRenderMethod)(this->EndRenderMethodArg);
    
}
